@@ -2,15 +2,22 @@
 import numpy as np 
 import h5py
 import os
+import multiprocessing
+import itertools
+import copy
+
+from scipy.interpolate import interp1d
 
 ## from abg_python
 from ..snapshot_utils import openSnapshot,get_unit_conversion
 from ..plot_utils import add_to_legend
 from ..color_utils import get_distinct
 from ..array_utils import findIntersection
-from ..snapshot_utils import getfinsnapnum
+from ..system_utils import getfinsnapnum
 from ..physics_utils import iterativeCoM
-from ..cosmo_utils import load_AHF,load_rockstar
+from ..cosmo_utils import load_AHF,load_rockstar,trace_rockstar
+from ..smooth_utils import smooth_x_varying_curve
+from ..math_utils import add_jhat_coords
 
 from .cosmoExtractor import extractDiskFromSnapdicts,offsetRotateSnapshot
 from .movie_utils import Draw_helper,FIREstudio_helper
@@ -19,6 +26,10 @@ from .metadata_utils import metadata_cache,Metadata,MultiMetadata
 from .firefly_utils import Firefly_helper
 from .scale_height_utils import ScaleHeight_helper
 
+try: from firestudio.utils.stellar_utils.load_stellar_hsml import get_particle_hsml
+except: print("FIRE studio is not installed, Galaxy.get_HSML will not work")
+
+
 ## mapping between particle type and what I called them
 sub_snap_dict = {
     0:'sub_snap',
@@ -111,9 +122,9 @@ class Galaxy(
             plot_color = 0 - color this instance will use when plotting onto an axis
             multi_thread = 1 - number of threads this instance will use if multi-threading
                 is available
-            ahf_path = None - path to AHF halo files, defaults first to my halo directory
+            halo_path = None - path to AHF halo files, defaults first to my halo directory
                 and snapdir/../halo/ahf if halo second.
-            ahf_fname = None - name of the AHF file, typically smooth_halo_00000.dat
+            halo_fname = None - name of the AHF file, typically smooth_halo_00000.dat
 
         Provided functions:
 
@@ -154,11 +165,10 @@ class Galaxy(
         snapdir_name = None,
         plot_color = 0,
         multi_thread = 1,
-        ahf_path = None,
-        ahf_fname = None,
         save_header_to_table = True,
         meta_name = None,
         suite_name = 'metal_diffusion',
+        use_rockstar_first=True,
         **metadata_kwargs 
         ):
 
@@ -346,30 +356,44 @@ class Galaxy(
             self.current_time_Gyr = self.header['TimeGyr']
 
         
-        if self.header['cosmological']:
-            ## opens the halo file to find the halo center and virial radius
-            self.load_halo_file()
-        else:
-            self.scom = np.zeros(3)
-            self.rvir = 300 ## what should i do here...
-            self.rstar_half = None
+            if self.header['cosmological']:
+                ## opens the halo file to find the halo center and virial radius
+                self.load_halo_file(use_rockstar_first=use_rockstar_first)
+            else:
+                self.scom = np.zeros(3)
+                self.rvir = 300 ## what should i do here...
+                self.rstar_half = None
 
-            if 'r30r' in self.snapdir:
-                self.scom = get_idealized_center(self.name,self.snapnum)
+                if 'r30r' in self.snapdir:
+                    self.scom = get_idealized_center(self.name,self.snapnum)
 
 
-        ## have we already calculated it and cached it?
-        if self.rstar_half is None:
-            for attr in ['gas_extract_rstar_half','star_extract_rstar_half']:
-                if hasattr(self.metadata,attr):
-                    self.rstar_half = getattr(self.metadata,attr)
-                    break
-
-            ## I guess not
+            ## have we already calculated it and cached it?
             if self.rstar_half is None:
-                print("No rstar 1/2 in halo or metadata files, we will need to calculate it ourselves.")
+                for attr in ['gas_extract_rstar_half','star_extract_rstar_half']:
+                    if hasattr(self.metadata,attr):
+                        self.rstar_half = getattr(self.metadata,attr)
+                        break
 
-    def load_halo_file(self,halo_fname=None,halo_path=None):
+                ## I guess not
+                if self.rstar_half is None:
+                    if self.metadata.loud_metadata:
+                        print("No rstar 1/2 in halo or metadata files, we will need to calculate it ourselves.")
+
+    def load_halo_file(
+        self,
+        halo_fname=None,
+        halo_path=None,
+        use_rockstar_first=True,
+        **kwargs):
+
+        ## decide which one is the fallback
+        if not use_rockstar_first:
+            first_fn = self.load_ahf
+            second_fn = self.load_rockstar
+        else:
+            first_fn = self.load_rockstar
+            second_fn = self.load_ahf
 
         if halo_path == 'None' or halo_fname == 'None':
             self.scom,self.rvir,self.rstar_half = None,None,None
@@ -380,37 +404,22 @@ class Galaxy(
                 'rstar_half',
                 'attributes manually')
         else:
-            try:
-                if 'elvis' in self.snapdir:
-                    raise IOError("No AHF files for Elvis runs")
-                self.load_ahf(ahf_fname=halo_fname,ahf_path=halo_path)
+            try: first_fn(halo_fname,halo_path,**kwargs)
             except IOError as e:
-                #print(e)
-                try:
-                    
-                    if 'elvis' in self.snapdir:
-                        which_host = self.snapdir.split('m12_elvis_')[1].split('_')[0]
-                        if which_host[:len(self.pretty_name)] == self.pretty_name:
-                            which_host = 0 
-                        else:
-                            if which_host[-len(self.pretty_name):] != self.pretty_name:
-                                raise IOError("invalid name, should be one of %s"%which_host)
-                            which_host = 1
-                    else:
-                        which_host = 0
-
-                    self.load_rockstar(
-                        rockstar_fname=halo_fname,
-                        rockstar_path=halo_path,
-                        which_host=which_host)
+                try: second_fn(halo_fname,halo_path,**kwargs)
                 except IOError:
                     print("Couldn't find AHF nor Rockstar halo files")
                     raise
 
-    def load_ahf(self,**kwargs):
+    def load_ahf(self,halo_fname=None,halo_path=None,**kwargs):
 
-        ## automatically search for the ahf_path and ahf_fname
-        self.ahf_path, self.ahf_fname = self.auto_search_ahf(**kwargs)
+        #raise Exception("We don't use ahf, only rockstar. join us.")
+
+        if 'elvis' in self.snapdir:
+            raise IOError("No AHF files for Elvis runs")
+
+        ## automatically search for the halo_path and halo_fname
+        self.halo_path, self.halo_fname = self.auto_search_ahf(**kwargs)
     
         ## now that we've attempted to identify an AHF file lets open 
         ##  this puppy up
@@ -419,9 +428,8 @@ class Galaxy(
             self.scom, self.rvir, self.rstar_half = load_AHF(
                 self.snapdir,
                 self.snapnum,
-                self.current_redshift,
-                ahf_path = self.ahf_path,
-                fname = self.ahf_fname,
+                ahf_path = self.halo_path,
+                fname = self.halo_fname,
                 hubble = self.header['HubbleParam'])
 
         except ValueError:
@@ -429,49 +437,48 @@ class Galaxy(
             self.scom, self.rvir = load_AHF(
                 self.snapdir,
                 self.snapnum,
-                self.current_redshift,
                 extra_names_to_read = [],
-                ahf_path = self.ahf_path,
-                fname = self.ahf_fname,
+                ahf_path = self.halo_path,
+                fname = self.halo_fname,
                 hubble = self.header['HubbleParam'])
 
             self.rstar_half = None
 
         return self.scom,self.rvir,self.rstar_half
             
-    def auto_search_ahf(self,ahf_fname=None,ahf_path=None):
+    def auto_search_ahf(self,halo_fname=None,halo_path=None):
         ## attempt to read halo location and properties from AHF
-        if ahf_fname is None:
-            ahf_fname='halo_0000%d_smooth.dat'%halo_id(self.snapdir_name)
+        if halo_fname is None:
+            halo_fname='halo_0000%d_smooth.dat'%halo_id(self.snapdir_name)
 
-        if ahf_path is None:
+        if halo_path is None:
             ## system blind if you put the soft link in
-            ahf_path = os.path.join(
+            halo_path = os.path.join(
                 os.environ['HOME'],'halo_files',
                 "%s","%s")
 
-            ahf_path = ahf_path%(self.suite_name,self.snapdir_name)
+            halo_path = halo_path%(self.suite_name,self.snapdir_name)
 
-        ## check if this first guess at the ahf_fname and ahf_path
+        ## check if this first guess at the halo_fname and halo_path
         ##  is right
-        if (not os.path.isfile(os.path.join(ahf_path,ahf_fname)) and 
-            ahf_path != 'None' and 
-            ahf_fname != 'None'):
+        if (not os.path.isfile(os.path.join(halo_path,halo_fname)) and 
+            halo_path != 'None' and 
+            halo_fname != 'None'):
             ## try looking in the simulation directory
-            ahf_path = os.sep.join(
+            halo_path = os.sep.join(
                 self.snapdir.split(os.sep)[:-1] ## remove output from the snapdir
                 +['halo','ahf']) ## look in <simdir>/halo/ahf
 
             ## okay no smooth halo file but there is a halo/ahf directory at least
-            if (os.path.isdir(ahf_path) and (
-                not os.path.isfile(os.path.join(ahf_path,ahf_fname)))):
+            if (os.path.isdir(halo_path) and (
+                not os.path.isfile(os.path.join(halo_path,halo_fname)))):
 
                 ## let's scan through the files in the directory and try and 
                 ##  find an AHF halo file that corresponds to just this snapshot. 
                 fnames = []
                 snap_string = "%03d"%self.snapnum
 
-                for fname in os.listdir(ahf_path):
+                for fname in os.listdir(halo_path):
                     if (snap_string in fname and 
                         'AHF_halos' in fname):
                         fnames+=[fname]
@@ -480,36 +487,180 @@ class Galaxy(
                 else:
                     raise IOError("can't find a halo file (or found too many)",fnames)
 
-        return ahf_path, ahf_fname
+        return halo_path, halo_fname
     
-    def get_ahf_file_output(self):
-        try: return load_AHF(
+    def get_rockstar_file_output(
+        self,
+        use_metadata=True,
+        save_meta=True,
+        loud=True,
+        assert_cached=False,
+        force_from_file=False,
+        fancy_trace=True,
+        smooth=None,#0.1, ## results in snapshots that are way mis-centered
+        **kwargs):
+        """fancy_trace=True - use additional phase space info to minimize jumps in rcom coordinates.
+            smooth=0.1 - whether to smooth by a window in Gyr, None for no smoothing"""
+
+        prefix = ''
+        if fancy_trace:
+            prefix+='fancy_'
+        if smooth is not None:
+            prefix+='smoothed%.3fGyr_'%smooth
+
+        ## separate cachefile that we can save to that all snapshots can use
+        hdf5_path = os.path.join(
+            os.environ['HOME'],
+            'halo_files',
+            'rockstar',
+            self.suite_name,
+            self.name,
+            prefix+'rockstar_trace.hdf5')
+
+        kwargs['smooth'] = smooth
+        kwargs['fancy_trace'] = fancy_trace
+
+        self.halo_path = os.path.dirname(hdf5_path)
+        self.halo_fname = os.path.basename(hdf5_path)
+            
+        @metadata_cache(
+            prefix+'rockstar_history',
+            [prefix+'snapnums',prefix+'rcoms',prefix+'rvirs'],
+            use_metadata=use_metadata,
+            save_meta=save_meta,
+            loud=loud,
+            assert_cached=assert_cached,
+            force_from_file=force_from_file)
+        def compute_rockstar_file_output(self,fancy_trace=True,smooth=0.1):
+
+            print(f'Tracing the rockstar halo files with fancy:{fancy_trace} and {smooth} Gyr smoothing.')
+            snapnums,rcoms,rvirs = trace_rockstar(self.snapdir,fancy_trace=fancy_trace)
+            new_rcoms = copy.copy(rcoms)
+            if smooth is not None:
+                self.get_snapshotTimes()
+                for i in range(3):
+                    smooth_times,smooth_coords,_,_,_ = smooth_x_varying_curve(
+                        self.snap_gyrs[-len(snapnums):],
+                        rcoms[:,i],
+                        smooth,
+                        assign='center')
+                    new_coords = interp1d(
+                        smooth_times,
+                        smooth_coords,
+                        bounds_error=False,
+                        fill_value=np.nan)(
+                        self.snap_gyrs[-len(snapnums):])
+                    new_rcoms[-len(snapnums):,i] = new_coords
+
+                    ## can't smooth the first and last half-window
+                    ##  /shrug
+                    outside_window_mask = np.isnan(new_rcoms[:,i])
+                    new_rcoms[outside_window_mask,i] = rcoms[outside_window_mask,i]
+                rcoms = new_rcoms
+
+            return snapnums,rcoms,rvirs
+
+        ## try loading from a cached trace
+        if os.path.isfile(hdf5_path) and use_metadata:
+            if loud: print("reading from",hdf5_path)
+            with h5py.File(hdf5_path,'r') as handle:
+                snapnums = handle[prefix+'rockstar_history'][prefix+'snapnums'][()]
+                rcoms = handle[prefix+'rockstar_history'][prefix+'rcoms'][()]
+                rvirs = handle[prefix+'rockstar_history'][prefix+'rvirs'][()]
+        ## trace the halo and save it if allowed
+        else:
+            snapnums,rcoms,rvirs = compute_rockstar_file_output(self,**kwargs)
+            if save_meta: self.metadata.export_to_file(hdf5_path,prefix+'rockstar_history',write_mode='w')
+        return snapnums,rcoms,rvirs
+    
+    def get_ahf_file_output(
+        self,
+        use_metadata=True,
+        save_meta=False,
+        loud=True,
+        assert_cached=False,
+        force_from_file=False,
+        **kwargs):
+        @metadata_cache(
+            'ahf_history',
+            ['snapnums','rcoms','rvirs'],
+            use_metadata=use_metadata,
+            save_meta=save_meta,
+            loud=loud,
+            assert_cached=assert_cached,
+            force_from_file=force_from_file)
+        def compute_ahf_file_output(self):
+            ## get the self.halo_path to point to rockstar if possible
+            self.load_ahf()
+            snapnums = load_AHF(
                 self.snapdir,
                 self.snapnum,
-                self.current_redshift,
-                ahf_path = self.ahf_path,
-                fname = self.ahf_fname,
-                hubble = self.header['HubbleParam'],
-                return_full_halo_file=True).T
-        except: return load_AHF(
-                self.snapdir,
-                self.snapnum,
-                self.current_redshift,
-                ahf_path = self.ahf_path,
-                fname = self.ahf_fname,
+                ahf_path = self.halo_path,
+                fname = self.halo_fname,
                 hubble = self.header['HubbleParam'],
                 return_full_halo_file=True,
-                extra_names_to_read=[]).T
+                extra_names_to_read=[])[:,1]
 
-    def load_rockstar(self,rockstar_fname=None,rockstar_path=None,which_host=0):
+            rcoms = np.zeros((snapnums.size,3))+np.nan
+            rvirs = np.zeros(snapnums.size)+np.nan
 
-        self.scom,self.rvir = load_rockstar(
-            self.snapdir,
-            self.snapnum,
-            fname=rockstar_fname,
-            rockstar_path=rockstar_path,
-            which_host=which_host)
-        self.rstar_half=None
+            for i in snapnums:
+                rcom,rvir = load_AHF(
+                    self.snapdir,
+                    i,
+                    ahf_path = self.halo_path,
+                    fname=self.halo_fname,
+                    hubble=self.header['HubbleParam'])[:2]
+
+                rcoms[int(i-snapnums[0]),:] = rcom
+                rvirs[int(i-snapnums[0])] = rvir
+
+            return snapnums[::-1],rcoms,rvirs
+        return compute_ahf_file_output(self,**kwargs)
+
+    def load_rockstar(self,halo_fname=None,halo_path=None,**kwargs):
+
+        ## have default kwargs but allow them to be ignored
+        ##  in most human readable way
+        if 'use_metadata' not in kwargs:
+            kwargs['use_metadata'] = True
+        if 'save_meta' not in kwargs:
+            kwargs['save_meta'] = True
+        if 'loud' not in kwargs:
+            kwargs['loud'] = False
+
+        try:
+            snapnums,rcoms,rvirs = self.get_rockstar_file_output(**kwargs)
+
+            index = np.argwhere(snapnums==self.snapnum).astype(int)[0][0]
+            self.scom = rcoms[index]
+            self.rvir = rvirs[index]
+            self.rstar_half = None
+
+        except IOError:
+            if 'elvis' in self.snapdir:
+                which_host = self.snapdir.split('m12_elvis_')[1].split('_')[0]
+                if which_host[:len(self.pretty_name)] == self.pretty_name:
+                    which_host = 0 
+                else:
+                    if which_host[-len(self.pretty_name):] != self.pretty_name:
+                        raise IOError("invalid name, should be one of %s"%which_host)
+                    which_host = 1
+            else:
+                which_host = 0
+            if halo_path is None:
+                self.halo_path = '../halo/rockstar_dm/'
+                self.halo_path = os.path.realpath(os.path.join(self.snapdir,self.halo_path))
+
+            self.halo_fname = 'halo_%03d.hdf5'%self.snapnum if halo_fname is None else halo_fname
+
+            self.scom,self.rvir = load_rockstar(
+                self.snapdir,
+                self.snapnum,
+                fname=halo_fname,
+                rockstar_path=halo_path,
+                which_host=which_host)
+            self.rstar_half=None
 
         return self.scom,self.rvir,self.rstar_half
 
@@ -535,6 +686,16 @@ class Galaxy(
 
         data_FIRE_SN_times = os.path.join(self.datadir,'%s.txt'%snaptimes)
 
+        if os.path.isfile(data_FIRE_SN_times):
+            (self.snapnums,
+                self.snap_sfs,
+                self.snap_zs,
+                self.snap_gyrs,
+                self.dTs) = np.genfromtxt(data_FIRE_SN_times,unpack=1)
+            return
+
+        data_FIRE_SN_times = os.path.join(self.snapdir,'..','%s.txt'%snaptimes)
+
         if os.path.isfile(data_FIRE_SN_times):
             (self.snapnums,
                 self.snap_sfs,
@@ -647,6 +808,9 @@ class Galaxy(
         overwrite_full_snaps_with_rotated_versions = False,
         free_mem = True, ## delete the full snapshot from RAM
         extract_DM = True, ## do we want the DM particles? 
+        compute_stellar_hsml=False,
+        loud=True,
+        jhat_coords=False,
         **kwargs):
         """
         radius = None -- radius of final sub_snap extraction, *not* orient_radius, 
@@ -675,7 +839,7 @@ class Galaxy(
             'rgas_half'],
             use_metadata=False,
             save_meta=save_meta,
-            loud=1)
+            loud=loud)
         def extract_halo_inner(
             self,
             orient_stars=True,
@@ -701,7 +865,8 @@ class Galaxy(
                 ##  rather than relying on the output of AHF
                 if self.rstar_half is None: self.get_rstar_half(
                     save_meta=save_meta,
-                    force_from_file=True)
+                    force_from_file=True,
+                    loud=False)
 
                 ## radius to calculate angular momentum
                 ##  to orient on 
@@ -744,7 +909,7 @@ class Galaxy(
                 try:
                     if not use_saved_subsnapshots: raise AssertionError("Told not to use saved sub-snapshots")
 
-                    print("Using the saved sub-snapshots for",self)
+                    if loud: print("Using the saved sub-snapshots for",self)
 
                     self.sub_snap = openSnapshot(
                         None,None,
@@ -798,7 +963,7 @@ class Galaxy(
                             self.scom - sub_scom,
                             self.scom , sub_scom)
 
-                    print("Successfully loaded a pre-extracted subsnap")
+                    if loud: print("Successfully loaded a pre-extracted subsnap")
 
                     ## pass these sub-snapshots into the rotation routine
                     which_snap = self.sub_snap
@@ -810,7 +975,7 @@ class Galaxy(
                     message = "Failed to open saved sub-snapshots"
                     #message+= ' %s'%error.__class__  
                     message+= ' %s'%repr(error)
-                    print(message)
+                    if loud: print(message)
 
                     ## have to load the full snapshots...
                     if 'snap' not in self.__dict__:
@@ -837,7 +1002,8 @@ class Galaxy(
                 dark_snap=which_dark_snap, ## dark_snap = None will ignore dark matter particles
                 orient_stars=orient_stars,
                 force_theta_TB=force_theta_TB,
-                force_phi_TB=force_phi_TB)
+                force_phi_TB=force_phi_TB,
+                loud=loud)
 
             ## unpack the return value
             if not extract_DM:
@@ -853,6 +1019,12 @@ class Galaxy(
             ##  oriented on stars or gas
             self.orient_stars = orient_stars
 
+            if compute_stellar_hsml and 'SmoothingLength' not in self.sub_star_snap: 
+                already_saved=False
+                self.sub_star_snap['SmoothingLength'] = self.get_HSML(
+                    loud=loud,
+                    save_meta=False)
+
             ## save for later, if requested
             if (use_saved_subsnapshots and
                 extract_DM and
@@ -862,6 +1034,7 @@ class Galaxy(
             if not hasattr(self,'rgas_half'):
                 self.rgas_half = self.calculate_half_mass_radius(
                     which_snap=self.sub_snap) 
+            
 
             return (self.sub_radius,
                 self.orient_stars,
@@ -894,7 +1067,18 @@ class Galaxy(
 
             if 'snap' in self.__dict__.keys():
                 del self.snap
-            print("Snapshot memory free")
+            if loud: print("Snapshot memory free")
+        
+        snapdicts = [self.sub_snap,self.sub_star_snap]
+        if extract_DM: snapdicts += [self.sub_dark_snap]
+
+        ## store a couple of things in the dictionary to identify it:
+        for snapdict in snapdicts:
+            snapdict['name'] = self.name
+            snapdict['snapnum'] = self.snapnum
+            snapdict['datadir'] = self.datadir
+
+            if jhat_coords: add_jhat_coords(snapdict)
 
         return return_value
 
@@ -919,8 +1103,53 @@ class Galaxy(
             return self.calculate_half_mass_radius(),
         return compute_rstar_half(self)
 
+    def get_HSML(
+        self,
+        snapdict_name='star',
+        use_metadata=True,
+        save_meta=True,
+        assert_cached=False,
+        loud=True,
+        **kwargs, 
+        ):
+        """Compute smoothing lengths for particles that don't have them,
+            typically collisionless particles (like stars). 
+
+            Input:
+
+                snapdict_name -- name in the form of `'%s_snapdict'%snapdict_name`
+                    that will be used to compute smoothing lengths for. 
+
+                use_metadata = True -- flag to search cache for result
+                save_meta = True -- flag to cache the result
+                assert_cached = False -- flag to require a cache hit
+                loud = True -- whether cache hits/misses should be announced
+                    to the console.
+                
+            Output:
+
+                smoothing_lengths -- numpy array of estimated smoothing lengths"""
+
+        @metadata_cache(
+            '%s_data'%snapdict_name,  ## hdf5 file group name
+            ['%s_SmoothingLengths'%snapdict_name],
+            use_metadata=use_metadata,
+            save_meta=save_meta,
+            assert_cached=assert_cached,
+            loud=loud,
+            force_from_file=True) ## read from cache file, not attribute of object
+        def compute_HSML(self,snapdict_name):
+            snapdict_name = ('sub_%s_snap'%snapdict_name).replace('sub_gas','sub')
+            snapdict = getattr(self,snapdict_name)
+            pos = snapdict['Coordinates']
+            smoothing_lengths = get_particle_hsml(pos[:,0],pos[:,1],pos[:,2])
+            return smoothing_lengths
+
+        return compute_HSML(self,snapdict_name,**kwargs)
+
     def load_stars(self,**kwargs):
-        print("Loading star particles of",self,'at',self.snapdir)
+        if self.metadata.loud_metadata:
+            print("Loading star particles of",self,'at',self.snapdir)
         if not hasattr(self,'star_snap'):
             self.star_snap = openSnapshot(
                 self.snapdir,
@@ -928,14 +1157,16 @@ class Galaxy(
                 **kwargs)
 
     def load_gas(self,**kwargs):
-        print("Loading gas particles of",self,'at',self.snapdir)
+        if self.metadata.loud_metadata:
+            print("Loading gas particles of",self,'at',self.snapdir)
         self.snap = openSnapshot(
             self.snapdir,
             self.snapnum,0,
             **kwargs)
 
     def load_dark_matter(self,**kwargs):
-        print("Loading dark matter particles of",self,'at',self.snapdir)
+        if self.metadata.loud_metadata:
+            print("Loading dark matter particles of",self,'at',self.snapdir)
         self.dark_snap = openSnapshot(
             self.snapdir,self.snapnum,1,
             **kwargs)
@@ -1010,9 +1241,12 @@ class Galaxy(
             ## put the common keys into ABG_Header
             ABG_Header = handle.create_group("ABG_Header")
             for key in common_keys:
+                if key == 'AngularMomentum': 
+                    print('why is AngularMomentum being output to ABG_Header in abg_python.galaxy.gal_utils?')
+                    continue
                 ABG_Header.attrs[key]=self.sub_snap[key]
 
-            derived_arrays = set(['Temperature','AgeGyr'])
+            derived_arrays = set(['Temperature','AgeGyr','AngularMomentum'])
             for ptype,extra_keys in zip(ptypes,extra_keyss):
                 abg_pgroup = ABG_Header.create_group('PartType%d'%ptype)
                 this_sub_snap = getattr(self,sub_snap_dict[ptype])
@@ -1170,6 +1404,35 @@ class Galaxy(
 ###### MANY GALAXY FUNCTIONS
 class ManyGalaxy(Galaxy):
     """ """
+    
+    def purge_metadata(
+        self,
+        group_name,
+        key_name=None,
+        snaplow=None,
+        snaphigh=None,
+        force=False,
+        loud=True,
+        mps=1):
+
+        if snaplow is None: snaplow = self.minsnap
+        if snaphigh is None: snaphigh = self.finsnap
+
+        argss = zip(
+            itertools.repeat(self),
+            np.arange(snaplow,snaphigh+1),
+            itertools.repeat(group_name),
+            itertools.repeat(key_name),
+            itertools.repeat(force),
+            itertools.repeat(loud))
+
+        if mps > 1:
+            with multiprocessing.Pool(multiprocessing.cpu_count()) as my_pool:
+                my_pool.starmap(purge_metadata_group_wrapper,argss)
+                my_pool.close()
+                my_pool.join()
+        else:
+            for args in argss: purge_metadata_group_wrapper(*args)
 
     def __repr__(self):
         return "%s many-galaxy wrapper"%(self.name)
@@ -1322,6 +1585,21 @@ class ManyGalaxy(Galaxy):
             setattr(self,'__getattr__',__getattr__)
             setattr(self,'__getitem__',__getitem__)
 
+    def get_final_orientation(self):
+        """Get Tait-Bryan xyz rotation angles from the final snapshot"""
+        last_galaxy = self.loadAtSnapshot(self.finsnap)
+        try:
+            ## attempt to read orientation from the cached metadata
+            phi_TB = last_galaxy.metadata.star_extract_phi_TB
+            theta_TB = last_galaxy.metadata.star_extract_theta_TB
+        except AttributeError:
+            ## extract the main halo and cache the output
+            last_galaxy.extractMainHalo(save_meta=True)
+            phi_TB = last_galaxy.metadata.star_extract_phi_TB
+            theta_TB = last_galaxy.metadata.star_extract_theta_TB
+
+        return theta_TB,phi_TB
+
     def find_galaxy_population(
         self,
         cursnap=600,
@@ -1372,8 +1650,8 @@ class ManyGalaxy(Galaxy):
             Options are:
                 plot_color=0,
                 multi_thread=1,
-                ahf_path=None,
-                ahf_fname=None,
+                halo_path=None,
+                halo_fname=None,
             """
 
         new_kwargs = {}
@@ -1400,4 +1678,15 @@ def get_idealized_center(savename,snapnum):
     try:
         return np.load(center_path)['centers'][snapnum]
     except:
-        return np.load(center_path)['centers'][snapnum-1]
\ No newline at end of file
+        return np.load(center_path)['centers'][snapnum-1]
+
+
+## MPS wrapper functions
+
+def purge_metadata_group_wrapper(many_galaxy,snapnum,group_name,key_name,force,loud):
+    galaxy = many_galaxy.loadAtSnapshot(snapnum)
+    galaxy.metadata.purge_metadata_group(
+        group_name=group_name,
+        key_name=key_name,
+        force=force,
+        loud=loud)
\ No newline at end of file
