@@ -22,7 +22,8 @@ class Metadata(object):
         sub_load_low_indices=None,
         sub_load_high_indices=None,
         target_last_sizes=None,
-        sub_load_exclude=None):
+        sub_load_exclude=None,
+        loud=True):
 
 
         self.metapath = metapath
@@ -75,7 +76,7 @@ class Metadata(object):
                     value = handle.attrs[key]
                     setattr(self,key,value)
         except IOError:
-            if self.loud_metadata:
+            if self.loud_metadata and loud:
                 print("Couldn't find a metadata file... for\n",self )
 
     def hasattr(self,attr):
@@ -143,10 +144,23 @@ class Metadata(object):
         else:
             setattr(self,key_to_save,value)
         
-    def save_to_metadata(self,group,key,value,mode='a',overwrite=0):
+    def save_to_metadata(self,group,key,value,mode='a',overwrite=0,attempts=0):
         ## first determine if this should be saved as an attribute of
         ##  the currently open metadata object
-        self.__save_to_metadata_object(group,key,value,overwrite)
+        if attempts <=0: self.__save_to_metadata_object(group,key,value,overwrite)
+        try: self.__save_to_metadata_file(group,key,value,mode,overwrite)
+        except OSError as e:
+            ## likely attempting to save to a metadata file that another thread
+            ##  has open. tsk tsk.
+            if attempts < 4:
+                ## let's wait a bit to try and de-sync from the other thread
+                ##  and then try again
+                time.sleep(np.random.randint(30,60))
+                self.save_to_metadata(group,key,value,mode,overwrite,attempts+1)
+            ## well, we tried 5 times so we might as well give up. 
+            else: raise e
+
+    def __save_to_metadata_file(self,group,key,value,mode,overwrite):
         with h5py.File(self.metapath,mode) as handle:
             if group == 'header':
                 if key not in handle.attrs.keys():
@@ -206,30 +220,62 @@ class Metadata(object):
             if self.loud_metadata:
                 print("Couldn't find a metadata file...",self.metapath)
 
-    def purge_metadata_group(self,group_name,loud=False,force=0):
+    def purge_metadata_group(self,group_name,key_name=None,loud=False,force=0):
         if not force:
             raise Exception("I'm sorry Dave, I'm afraid I can't do that.")
         else:
             with h5py.File(self.metapath,'a') as handle:
                 if group_name == 'Header':
-                    for key in handle.attrs.keys():
-                        del handle.attrs[key]
-                if group_name in handle.keys():
+                    if key_name is None:
+                        ## delete all the keys
+                        for key in handle.attrs.keys():
+                            del handle.attrs[key]
+                    ## delete only the specified key
+                    else: del handle.attrs[key_name]
+                    return
+                
+                ## take advantage of h5py group syntax to delete only
+                ##  the specified key
+                if key_name is not None: group_name = "%s/%s"%(group_name,key_name)
+
+                if group_name in handle:
                     if self.loud_metadata or loud:
-                        print(handle.keys(),'...before...')
+                        print(np.array([
+                            (key,len(handle[key])) for key in handle.keys()]),
+                            '...before...')
                     del handle[group_name]
                     if self.loud_metadata or loud:
-                        print(handle.keys(),"...after. I hope you're happy.")
+                        print(np.array([
+                            (key,len(handle[key])) for key in handle.keys()]),
+                            "...after. I hope you're happy.")
                 else:
                     if self.loud_metadata or loud:
                         print("This metadata doesn't have %s."%group_name)
 
-            ## now get it out of the instance itself
-            for key in list(self.__dict__.keys()):
-                if key[:len(group_name)] == group_name:
-                    self.__dict__.pop(key)
+            if key_name is not None: 
+                group_name = group_name.replace('/','_')
+                if group_name in self.__dict__.keys():
+                    self.__dict__.pop(key_name)
+            else:
+                ## now get the whole group out of the instance
+                for key in list(self.__dict__.keys()):
+                    if key[:len(group_name)] == group_name:
+                        self.__dict__.pop(key)
 
-    def lazy_load_from_file(self,key,load_entire_group=False):
+    def lazy_load_from_file(self,key,load_entire_group=False,attempts=0):
+        try: self.__lazy_load_from_file(key,load_entire_group=load_entire_group)
+        except OSError as e: 
+            ## likely attempting to load from a metadata file that another thread
+            ##  has open. tsk tsk.
+            if attempts < 4:
+                ## let's wait a bit to try and de-sync from the other thread
+                ##  and then try again
+                time.sleep(np.random.randint(30,60))
+                self.lazy_load_from_file(key,load_entire_group,attempts+1)
+            ## well, we tried 5 times so we might as well give up. 
+            else: raise e
+    
+    def __lazy_load_from_file(self,key,load_entire_group=False):
         if key in self.file_keys:
             with h5py.File(self.metapath,'r') as handle:
                 ## handle groups
@@ -289,20 +335,6 @@ class Metadata(object):
                 return np.array(handle[pathh])
         else:
             return np.array(handle[pathh])
-    
-    def purge_metadata_key(self,group_name,key_name,force=0,loud=True):
-        if not force:
-            raise Exception("I'm sorry Dave, I'm afraid I can't do that.")
-        else:
-            with h5py.File(self.metapath,'a') as handle:
-                if group_name in handle.keys() and key_name in handle[group_name].keys():
-                    if self.loud_metadata and loud:
-                        print(handle[group_name].keys(),'...before...')
-                    del handle[group_name][key_name]
-                    if self.loud_metadata and loud:
-                        print(handle[group_name].keys(),"...after. I hope you're happy.")
-                else:
-                    raise IOError("%s doesn't have %s/%s."%(repr(self),group_name,key))
 
     def export_to_file(self,target,group_name,keys=None,write_mode='a'):
         if type(target) == dict:
@@ -312,6 +344,9 @@ class Metadata(object):
         if target[0] != os.sep:
             target = os.path.join(os.environ['HOME'],target)
             print('Detected relative path, prepending $HOME, new target:',target)
+        
+        dirname = os.path.dirname(target)
+        if not os.path.isdir(dirname): os.makedirs(dirname)
 
         ## make a new hdf5 file
         with h5py.File(target,write_mode) as target_handle:
@@ -446,10 +481,10 @@ def metadata_cache(
             if clean_metadata:
                 print('Clearing metadata of %s and exiting.'%func_name)
                 for key in keys:
-                    self.metadata.purge_metadata_key(
+                    self.metadata.purge_metadata_group(
                         group,
                         key,
-                        force=force,
+                        force=True,
                         loud=False)
                 return
                     
