@@ -13,7 +13,7 @@ from .array_utils import pairFilter
 from .fitting_utils import covarianceEigensystem
 from scipy.interpolate import interp1d
 
-latex_pagewidth=6.9738480697 ## in #7.125
+latex_pagewidth=6.9738480697 ## in
 latex_columnwidth=3.32 ## in
 
 
@@ -81,7 +81,21 @@ def add_many_to_legend(
 
 
 
-    
+
+def detach_legend(ax,legend=None,**legend_kwargs):
+    if legend is None:
+        legend = ax.get_legend()
+
+    ## if it's still None then there's not a legend
+    if legend is None:
+        return
+
+    lines = legend.get_lines()
+    labels = [text.get_text() for text in legend.get_texts()]
+
+    ## clear out the current legend
+    ax.legend([])
+    ax.legend(lines,labels,**legend_kwargs)
         
 def add_to_legend(
     ax,
@@ -710,7 +724,7 @@ def nameAxes(
         #ax.xaxis.set_minor_formatter(my_log_ticker))
         ax.xaxis.set_minor_formatter(NullFormatter())
     if logflag[1] :
-        ax.set_yscale('log',nonposy='clip')
+        ax.set_yscale('log',nonpositive='clip')
         ax.yaxis.set_major_formatter(my_log_ticker)
         #ax.yaxis.set_minor_formatter(my_log_ticker))
         ax.yaxis.set_minor_formatter(NullFormatter())
@@ -1063,7 +1077,7 @@ def set_matplotlib_params_black_talk(matplotlib):
 
 def ffmpeg_frames(
     framedir,
-    frame_heads,
+    frame_patterns,
     savename=None,
     framerate=15,
     extension='.mp4',
@@ -1075,10 +1089,12 @@ def ffmpeg_frames(
     if outdir is None:
         outdir = os.path.dirname(framedir)
     
-    for frame_head in frame_heads:
+    for frame_pattern in frame_patterns:
+        ## assume the frame pattern is something like "frame_%03d.png"
+        frame_head = frame_pattern.split('%')[0][:-1]
+
         cmd = 'ffmpeg -framerate %d'%framerate
-        cmd += ' -i %s'%os.path.join(framedir,frame_head)
-        cmd += '_%03d.png'
+        cmd += ' -i %s'%os.path.join(framedir,frame_pattern)
         cmd += ' -q:v 1'
         cmd += ' %s%s -y'%(os.path.join(outdir,frame_head),extension)
         print(cmd)
@@ -1110,3 +1126,55 @@ def place_text(ax,percentage):
     if logflag: 
         foo = 10**foo
     return foo
+
+try: import palettable
+except: print("palettable colormaps are not installed")
+try:
+    from pfh_colormaps import load_my_custom_color_tables
+    load_my_custom_color_tables()
+except:
+    print("don't have phil's colormaps")
+
+def get_cmap(cmap_name):
+    try: cmap = plt.get_cmap(cmap_name)
+    except:
+        ## perhaps i was passed a palettable cmap path
+        ## split the cmap path into its components
+        cmap_name = cmap_name.split(".")
+
+        ## iteratively getattr to achieve zen
+        cmap = palettable
+        for name in cmap_name: cmap = getattr(cmap,name)
+
+        ## extract the matplotlib colormap object
+        cmap = cmap.mpl_colormap
+    return cmap
+
+
+### from https://kavigupta.org/2019/05/18/Setting-the-size-of-figures-in-matplotlib/?s=03
+from matplotlib.image import imread
+from tempfile import NamedTemporaryFile
+
+def get_size(fig, dpi=240):
+    with NamedTemporaryFile(suffix='.png') as f:
+        fig.savefig(f.name, bbox_inches='tight',dpi=dpi)
+        height, width, _channels = imread(f.name).shape
+        return width/dpi , height/dpi
+
+def set_size(fig, size, eps=1e-2, give_up=2, min_size_in=0.25):
+
+    target_width, target_height = size
+    set_width, set_height = target_width, target_height # reasonable starting point
+    deltas = [] # how far we have
+    while True:
+        fig.set_size_inches([set_width, set_height])
+        actual_width, actual_height = get_size(fig)
+        set_width *= target_width / actual_width
+        set_height *= target_height / actual_height
+        deltas.append(abs(actual_width - target_width) + abs(actual_height - target_height))
+        if deltas[-1] < eps:
+            return True
+        if len(deltas) > give_up and sorted(deltas[-give_up:]) == deltas[-give_up:]:
+            return False
+        if set_width < min_size_in or set_height < min_size_in:
+            return False
\ No newline at end of file
